#ifndef TCL_HPDF_H
#define TCL_HPDF_H
#include <tcl.h>
#include <hpdf.h>
#include <hpdf_conf.h>
#include <hpdf_utils.h>
#include <hpdf_pages.h>
#include <string.h>

#define TCLHPDF_ARGCHECK(a,b,c,d)                   \
if (objc < (a) || (b) < objc) {                     \
    Tcl_WrongNumArgs(interp, (c), objv, (d));       \
    return TCL_ERROR;                               \
}

#define hpdf_ErrorReturn(interp, pdf)               \
if (HPDF_GetError(pdf) != HPDF_OK) {                \
    if (strlen(Tcl_GetStringResult(interp)) == 0) { \
        Tcl_ResetResult(interp);                    \
        Tcl_AppendResult(interp,                    \
        hpdf_getErrorMessage(HPDF_GetError(pdf)),   \
        (char*)NULL);                               \
    }                                               \
    HPDF_ResetError(pdf);                           \
    return TCL_ERROR;                               \
}

#define hpdf_ErrorReturn2(interp, pdf,ret)          \
if (ret != HPDF_OK) {                               \
    if (strlen(Tcl_GetStringResult(interp)) == 0) { \
        Tcl_ResetResult(interp);                    \
        Tcl_AppendResult(interp,                    \
        hpdf_getErrorMessage(ret),                  \
        (char*)NULL);                               \
    }                                               \
    HPDF_ResetError(pdf);                           \
    return TCL_ERROR;                               \
}

#ifdef _WIN32
#define snprintf _snprintf
#endif

/*
 * tclhpdf_obj.c
 */
typedef enum {
    TCLHPDF_TYPE_DOC,
    TCLHPDF_TYPE_PAGE,
    TCLHPDF_TYPE_IMAGE,
    TCLHPDF_TYPE_FONT,
    TCLHPDF_TYPE_OUTLINE,
    TCLHPDF_TYPE_ENCODER,
    TCLHPDF_TYPE_DESTINATION,
    TCLHPDF_TYPE_ANNOTATION,
    TCLHPDF_TYPE_EXTGSTATE
} hpdf_ObjType;

typedef struct hpdf_Obj {
    hpdf_ObjType    type;
    void            *data;
    Tcl_Interp      *interp;
    Tcl_Command     cmdtoken;

    struct hpdf_Obj *root;
    struct hpdf_Obj *parent;
    struct hpdf_Obj *children;
    struct hpdf_Obj *next;
    struct hpdf_Obj *prev;
} hpdf_Obj;

extern int hpdf_NewHpdfCmd(Tcl_Interp *interp, hpdf_ObjType type, hpdf_Obj *parent, void* data);
extern hpdf_Obj* hpdf_ObjAlloc (hpdf_ObjType type, hpdf_Obj *parent, 
        Tcl_Interp *interp, void* data);
extern void hpdf_ObjDestroy (hpdf_Obj *obj, int removeCmdFlag);
extern int hpdf_ObjValidateType(hpdf_Obj *obj, hpdf_ObjType type);
extern int hpdf_GetObjFromCmdName (Tcl_Interp *interp, 
        Tcl_Obj *nameobj, hpdf_ObjType type, hpdf_Obj **obj);
#define TCLHPDFOBJ_GET_DOC(obj) ((HPDF_Doc)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_PAGE(obj) ((HPDF_Page)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_IMAGE(obj) ((HPDF_Image)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_FONT(obj) ((HPDF_Font)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_OUTLINE(obj) ((HPDF_Outline)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_ENCODER(obj) ((HPDF_Encoder)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_DESTINATION(obj) ((HPDF_Destination)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_ANNOTATION(obj) ((HPDF_Annotation)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_EXTGSTATE(obj) ((HPDF_ExtGState)((hpdf_Obj*)(obj))->data)
#define TCLHPDFOBJ_GET_ROOTDOC(obj) ((HPDF_Doc)((hpdf_Obj*)(obj))->parent->data)

int hpdf_TclListToHpdfRect (Tcl_Interp *interp, const char *name, Tcl_Obj *list, HPDF_Rect *rect);
char * hpdf_TclObjToHpdfText (Tcl_Obj *obj, Tcl_DString *ds);

/*
 * tclhpdf.c
 */
extern const char *hpdf_getErrorMessage (HPDF_STATUS error);
/*
 * tclhpdf_doc.c
 */
extern int hpdf_DocObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
void hpdf_DocCmdDeleteProc(ClientData clientData);
/*
 * tclhpdf_page.c
 */
extern int hpdf_PageObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_image.c
 */
extern int hpdf_ImageObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_font.c
 */
extern int hpdf_FontObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_outline.c
 */
extern int hpdf_OutlineObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_encoder.c
 */
extern int hpdf_EncoderObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_destination.c
 */
extern int hpdf_DestinationObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_annotation.c
 */
extern int hpdf_AnnotationObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
/*
 * tclhpdf_extgstate.c
 */
extern int hpdf_ExtGStateObjCmd(ClientData clientData, 
        Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);


/*
 * tclhpdf_const.c
 */
extern const char *optionCompressionModeString[];
extern const int optionCompressionModeEnum[];
extern const char *optionPermissionString[];
extern const int optionPermissionEnum[];
extern const char *optionViewerPreferenceString[];
extern const int optionViewerPreferenceEnum[];
extern const char *optionGraphicsModeString[];
extern const int optionGraphicsModeEnum[];

extern const char *optionInfoTypeGetString[];
extern const HPDF_InfoType optionInfoTypeGetEnum[];
extern const char *optionInfoTypeSetString[];
extern const HPDF_InfoType optionInfoTypeSetEnum[];
extern const char *optionInfoTypeDateString[];
extern const HPDF_InfoType optionInfoTypeDateEnum[];
extern const char *optionPDFVerString[];
extern const HPDF_PDFVer optionPDFVerEnum[];
extern const char *optionEncryptModeString[];
extern const HPDF_EncryptMode optionEncryptModeEnum[];
extern const char *optionColorSpaceString[];
extern const HPDF_ColorSpace optionColorSpaceEnum[];
extern const char *optionColorSpaceForRawString[];
extern const HPDF_ColorSpace optionColorSpaceForRawEnum[];
extern const char *optionLineCapString[];
extern const HPDF_LineCap optionLineCapEnum[];
extern const char *optionLineJoinString[];
extern const HPDF_LineJoin optionLineJoinEnum[];
extern const char *optionTextRenderingModeString[];
extern const HPDF_TextRenderingMode optionTextRenderingModeEnum[];
extern const char *optionWritingModeString[];
extern const HPDF_WritingMode optionWritingModeEnum[];
extern const char *optionPageLayoutString[];
extern const HPDF_PageLayout optionPageLayoutEnum[];
extern const char *optionPageModeString[];
extern const HPDF_PageMode optionPageModeEnum[];
extern const char *optionPageNumStyleString[];
extern const HPDF_PageNumStyle optionPageNumStyleEnum[];
extern const char *optionDestinationTypeString[];
extern const HPDF_DestinationType optionDestinationTypeEnum[];
extern const char *optionAnnotTypeString[];
extern const HPDF_AnnotType optionAnnotTypeEnum[];
extern const char *optionAnnotFlgsString[];
extern const HPDF_AnnotFlgs optionAnnotFlgsEnum[];
extern const char *optionAnnotHighlightModeString[];
extern const HPDF_AnnotHighlightMode optionAnnotHighlightModeEnum[];
extern const char *optionAnnotIconString[];
extern const HPDF_AnnotIcon optionAnnotIconEnum[];
extern const char *optionBSSubtypeString[];
extern const HPDF_BSSubtype optionBSSubtypeEnum[];
extern const char *optionBlendModeString[];
extern const HPDF_BlendMode optionBlendModeEnum[];
extern const char *optionTransitionStyleString[];
extern const HPDF_TransitionStyle optionTransitionStyleEnum[];
extern const char *optionPageSizesString[];
extern const HPDF_PageSizes optionPageSizesEnum[];
extern const char *optionPageDirectionString[];
extern const HPDF_PageDirection optionPageDirectionEnum[];
extern const char *optionEncoderTypeString[];
extern const HPDF_EncoderType optionEncoderTypeEnum[];
extern const char *optionByteTypeString[];
extern const HPDF_ByteType optionByteTypeEnum[];
extern const char *optionTextAlignmentString[];
extern const HPDF_TextAlignment optionTextAlignmentEnum[];

extern int hpdf_getStringObjFromEnum (Tcl_Interp *interp,
        const char *strings[], const int enums[],
        int val, Tcl_Obj **result);
extern int hpdf_getEnumFromStringObj (Tcl_Interp *interp,
        const char *strings[], const int enums[],
        Tcl_Obj *val, int *result);
extern int hpdf_getListObjFromFlags (Tcl_Interp *interp,
        const char *strings[], const int enums[],
        int val, Tcl_Obj **result);
extern int hpdf_getFlagsFromListObj (Tcl_Interp *interp,
        const char *strings[], const int enums[],
        Tcl_Obj *listObj, int *result);


#endif
