#include "tclhpdf.h"

int hpdf_FontObjCmd(ClientData clientData, Tcl_Interp *interp, 
        int objc, Tcl_Obj *const objv[])
{
    int index;
    HPDF_Font font = TCLHPDFOBJ_GET_FONT(clientData);
    HPDF_Doc pdf = TCLHPDFOBJ_GET_ROOTDOC(clientData);

    static CONST char *optionStrings[] = {
    "getFontName",
    "getEncodingName",
    "getUnicodeWidth",
    "getBBox",
    "getAscent",
    "getDescent",
    "getXHeight",
    "getCapHeight",
    "textWidth",
    "measureText",
     (char*)NULL
    };

    enum options {
    TCLHPDF_Font_GetFontName,
    TCLHPDF_Font_GetEncodingName,
    TCLHPDF_Font_GetUnicodeWidth,
    TCLHPDF_Font_GetBBox,
    TCLHPDF_Font_GetAscent,
    TCLHPDF_Font_GetDescent,
    TCLHPDF_Font_GetXHeight,
    TCLHPDF_Font_GetCapHeight,
    TCLHPDF_Font_TextWidth,
    TCLHPDF_Font_MeasureText
    };

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
        return TCL_ERROR;
    }
    if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, 
                "option", 0, &index) != TCL_OK) {
        return TCL_ERROR;
    }

    switch (index) {
        case TCLHPDF_Font_GetFontName:
        {
            const char *name;
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            name = HPDF_Font_GetFontName(font);

            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewStringObj(name, -1));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetEncodingName:
        {
            const char *name;
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            name = HPDF_Font_GetEncodingName(font);

            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewStringObj(name, -1));
            return TCL_OK;
        }
            
        case TCLHPDF_Font_GetUnicodeWidth:
        {
            int width;
            Tcl_UniChar tclunicode;
            TCLHPDF_ARGCHECK(3,3,2,"unichar")

            tclunicode = Tcl_GetUniChar(objv[2], 0);
            width = HPDF_Font_GetUnicodeWidth(font, (HPDF_UNICODE)tclunicode);

            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewIntObj(width));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetBBox:
        {
            Tcl_Obj *objrect[4];
            HPDF_Box rect;
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            rect = HPDF_Font_GetBBox(font);
            hpdf_ErrorReturn(interp, pdf);
            
            objrect[0] = Tcl_NewIntObj(rect.left);
            objrect[1] = Tcl_NewIntObj(rect.bottom);
            objrect[2] = Tcl_NewIntObj(rect.right);
            objrect[3] = Tcl_NewIntObj(rect.top);
            Tcl_SetObjResult(interp, Tcl_NewListObj(4, objrect));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetAscent:
        {
            HPDF_INT i;
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            i = HPDF_Font_GetAscent(font);
            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewIntObj(i));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetDescent:
        {
            HPDF_INT i;
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            i = HPDF_Font_GetDescent(font);
            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewIntObj(i));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetXHeight:
        {
            HPDF_INT i;
            i = HPDF_Font_GetXHeight(font);
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewIntObj(i));
            return TCL_OK;
        }

        case TCLHPDF_Font_GetCapHeight:
        {
            HPDF_INT i;
            i = HPDF_Font_GetCapHeight(font);
            TCLHPDF_ARGCHECK(2,2,2,NULL)

            hpdf_ErrorReturn(interp, pdf);
            Tcl_SetObjResult(interp, Tcl_NewIntObj(i));
            return TCL_OK;
        }

        case TCLHPDF_Font_TextWidth:
        {
            Tcl_DString ds;
            Tcl_Obj *twobj[4];
            HPDF_TextWidth textwidth;
            const char *text;
            int len;
            TCLHPDF_ARGCHECK(4,4,2,"text len")
            
            if (Tcl_GetIntFromObj(interp, objv[3], &len) != TCL_OK) {
                return TCL_ERROR;
            }
            
            if (objv[2]->typePtr && strcmp(objv[2]->typePtr->name, "bytearray")) {
                text = hpdf_TclObjToHpdfText(objv[2], &ds);
                if (strlen(text) < len) {len = strlen(text);}
                textwidth = HPDF_Font_TextWidth(font, text, len);
            } else {
                text = hpdf_TclObjToHpdfText(Tcl_GetRange(objv[2], 0, len-1), &ds);
                textwidth = HPDF_Font_TextWidth(font, text, strlen(text));
            }
            Tcl_DStringFree(&ds);

            hpdf_ErrorReturn(interp, pdf);
            
            twobj[0] = Tcl_NewIntObj(textwidth.numchars);
            twobj[1] = Tcl_NewIntObj(textwidth.numwords);
            twobj[2] = Tcl_NewIntObj(textwidth.width);
            twobj[3] = Tcl_NewIntObj(textwidth.numspace);
            
            Tcl_SetObjResult(interp, Tcl_NewListObj(4, twobj));
            Tcl_DStringFree(&ds);
            return TCL_OK;
        }

        case TCLHPDF_Font_MeasureText:
        {
            Tcl_DString ds;
            const char *text;
            int len, wordwrap, blen, result;
            double width, fontsize, charspace, wordspace;
            HPDF_REAL realwidth;
            TCLHPDF_ARGCHECK(9,10,2,"text len width fontsize charspace wordspace wordwrap ?realwidthName?")
            
            if (Tcl_GetIntFromObj(interp, objv[3], &len) != TCL_OK) {
                return TCL_ERROR;
            }

            text = hpdf_TclObjToHpdfText(Tcl_GetRange(objv[2], 0, len-1), &ds);

            if (Tcl_GetDoubleFromObj(interp, objv[4], &width) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Tcl_GetDoubleFromObj(interp, objv[5], &fontsize) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Tcl_GetDoubleFromObj(interp, objv[6], &charspace) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Tcl_GetDoubleFromObj(interp, objv[7], &wordspace) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Tcl_GetBooleanFromObj(interp, objv[8], &wordwrap) != TCL_OK) {
                return TCL_ERROR;
            }
            
            if (objc == 10) {
                blen = HPDF_Font_MeasureText (font, text, len, width, 
                        fontsize, charspace, wordspace, wordwrap, &realwidth);
            } else {
                blen = HPDF_Font_MeasureText (font, text, len, width, 
                        fontsize, charspace, wordspace, wordwrap, NULL);
            }
            hpdf_ErrorReturn(interp, pdf);

            if (objc == 10) {
                Tcl_Obj *ret;
                ret = Tcl_ObjSetVar2(interp, objv[9], NULL, 
                        Tcl_NewDoubleObj(realwidth), 0);
                if (ret == NULL) 
                    return TCL_ERROR;
            }
            
            {
                Tcl_DString ds;
                Tcl_DStringInit(&ds);
                result = Tcl_GetCharLength(Tcl_NewStringObj(Tcl_ExternalToUtfDString(
                            NULL, text, blen, &ds),-1));
                Tcl_DStringFree(&ds);
            }
            Tcl_SetObjResult(interp, Tcl_NewIntObj(result));
            return TCL_OK;
        }

        default: break;
    }
    return TCL_OK;
}
