#include "tclhpdf.h"

static void set_child (hpdf_Obj *parent, hpdf_Obj *child);
static void destroy_children (hpdf_Obj *obj);
static hpdf_Obj* get_last_child (hpdf_Obj *obj);
static hpdf_Obj* get_first_child (hpdf_Obj *obj);
static hpdf_Obj* append_first (hpdf_Obj *list, hpdf_Obj *obj);
static hpdf_Obj* append_last  (hpdf_Obj *list, hpdf_Obj *obj);


int hpdf_NewHpdfCmd(Tcl_Interp *interp, hpdf_ObjType type, hpdf_Obj *parent, void* data)
{
    char name[256];
    hpdf_Obj *obj;
    char *format;
    void *proc = NULL;
    void *delproc = NULL;

    switch (type) {
        case TCLHPDF_TYPE_DOC:
            format = "hpdfDoc%X";
            proc = hpdf_DocObjCmd;
            delproc = hpdf_DocCmdDeleteProc;
            break;
        case TCLHPDF_TYPE_PAGE:
            format = "hpdfPage%X";
            proc = hpdf_PageObjCmd;
            break;
        case TCLHPDF_TYPE_IMAGE:
            format = "hpdfImage%X";
            proc = hpdf_ImageObjCmd;
            break;
        case TCLHPDF_TYPE_FONT:
            format = "hpdfFont%X";
            proc = hpdf_FontObjCmd;
            break;
        case TCLHPDF_TYPE_OUTLINE:
            format = "hpdfOutline%X";
            proc = hpdf_OutlineObjCmd;
            break;
        case TCLHPDF_TYPE_ENCODER:
            format = "hpdfEncoder%X";
            proc = hpdf_EncoderObjCmd;
            break;
        case TCLHPDF_TYPE_DESTINATION:
            format = "hpdfDestination%X";
            proc = hpdf_DestinationObjCmd;
            break;
        case TCLHPDF_TYPE_ANNOTATION:
            format = "hpdfAnnotation%X";
            proc = hpdf_AnnotationObjCmd;
            break;
        case TCLHPDF_TYPE_EXTGSTATE:
            format = "hpdfExtGState%X";
            proc = hpdf_ExtGStateObjCmd;
            break;
        default:
            snprintf(name, 256, "faild to create a new hpdf objcmd. type(%d)", type);
            Tcl_SetObjResult(interp, Tcl_NewStringObj(name,-1));
            return TCL_ERROR;
    }


    obj = hpdf_ObjAlloc(type, parent, interp, data);
    
    snprintf(name, 256, format, (unsigned int)data);

    obj->cmdtoken = Tcl_CreateObjCommand(interp, name, 
        (Tcl_ObjCmdProc *)proc,
        (ClientData)obj, 
        (Tcl_CmdDeleteProc *)delproc);

    Tcl_SetObjResult(interp, Tcl_NewStringObj(name, -1));
    return TCL_OK;
}

hpdf_Obj* hpdf_ObjAlloc (hpdf_ObjType type, hpdf_Obj *parent, Tcl_Interp *interp, void* clientData)
{
    hpdf_Obj* obj;
    obj = (hpdf_Obj*)Tcl_Alloc(sizeof(hpdf_Obj));
    obj->type      = type;
    obj->data      = clientData;
    obj->interp    = interp;
    obj->cmdtoken  = NULL;
    if (parent) {
    obj->root      = parent->root;
    obj->parent    = parent;
    } else {
    obj->root      = obj;
    obj->parent    = NULL;
    }
    obj->children  = NULL;
    obj->next      = NULL;
    obj->prev      = NULL;

    if (parent) {
        set_child (parent, obj);
    }
    return obj;
}

void hpdf_ObjDestroy (hpdf_Obj *obj, int removeCmdFlag)
{
    if (obj == NULL) return;
    
    destroy_children (obj);

    if (obj->prev || obj->next) {
        hpdf_Obj *prev = obj->prev;
        hpdf_Obj *next = obj->next;
        if (prev) prev->next = next;
        if (next) next->prev = prev;
    }

    if (obj->type == TCLHPDF_TYPE_DOC) {
        HPDF_Free((HPDF_Doc)obj->data);
        obj->data = NULL;
    }

    if (removeCmdFlag && obj->cmdtoken) {
        const char *cmdname;
        cmdname = Tcl_GetCommandName(obj->interp, obj->cmdtoken);
        if (cmdname) {
            Tcl_DeleteCommand(obj->interp, cmdname);
        }
    }
    Tcl_Free((void*)obj);
}

int hpdf_ObjValidateType(hpdf_Obj *obj, hpdf_ObjType type)
{
    Tcl_CmdInfo info; 

    if (obj == NULL || obj->type != type ||
        obj->cmdtoken == NULL || obj->interp == NULL) 
        return TCL_ERROR;

    if (Tcl_GetCommandInfoFromToken(obj->cmdtoken, &info) != 1) {
        return TCL_ERROR;
    }
    return TCL_OK;
}


int hpdf_GetObjFromCmdName (Tcl_Interp *interp, 
        Tcl_Obj *nameobj, hpdf_ObjType type, hpdf_Obj **obj)
{
    char *name;
    Tcl_CmdInfo info; 

    name = Tcl_GetString(nameobj);

    if (Tcl_GetCommandInfo(interp, name, &info) == 0) {
        Tcl_AppendResult(interp, "unknown command name: ", name, NULL);
        return TCL_ERROR;
    }
    
    if (hpdf_ObjValidateType((hpdf_Obj*)info.objClientData, type) != TCL_OK) {
        char *typename;
        switch (type) {
            case TCLHPDF_TYPE_DOC:        typename = "hpdfDocObj"; break;
            case TCLHPDF_TYPE_PAGE:       typename = "hpdfPageObj"; break;
            case TCLHPDF_TYPE_IMAGE:      typename = "hpdfImageObj"; break;
            case TCLHPDF_TYPE_FONT:       typename = "hpdfFontObj"; break;
            case TCLHPDF_TYPE_OUTLINE:    typename = "hpdfOutlineObj"; break;
            case TCLHPDF_TYPE_ENCODER:    typename = "hpdfEncoderObj"; break;
            case TCLHPDF_TYPE_DESTINATION:typename = "hpdfDestinationObj"; break;
            case TCLHPDF_TYPE_ANNOTATION: typename = "hpdfAnnotationObj"; break;
            case TCLHPDF_TYPE_EXTGSTATE:  typename = "hpdfExtGStateObj"; break;
            default: typename = "hpdfUnknownObj"; break;
        }
        Tcl_AppendResult(interp, "the command is not ",
                typename, ": ", name, NULL);
        return TCL_ERROR;
    }
    *obj = info.objClientData;
    return TCL_OK;
}

/*
 * static
 */

static void set_child (hpdf_Obj *parent, hpdf_Obj *child)
{
    if (!parent || !child) return;
    
    if (parent->children == NULL) {
        parent->children = child;
    } else {
        parent->children = append_first (parent->children, child);
    }
    child->root   = parent->root;
    child->parent = parent;
    return;
}

static void destroy_children (hpdf_Obj *obj)
{
    hpdf_Obj *c;
    hpdf_Obj *next;
    if (obj->children) {
        c = get_first_child(obj->children);
        while (c != NULL) {
            next = c->next;
            hpdf_ObjDestroy(c, 1);
            c = next;
        }
    }
    obj->children = NULL;
}

static hpdf_Obj* append_first (hpdf_Obj *list, hpdf_Obj *obj)
{
    if (!list) return obj;

    list = get_first_child(list);
    obj->prev = NULL;
    obj->next = list;
    list->prev = obj;
    return obj;
}

static hpdf_Obj* append_last (hpdf_Obj *list, hpdf_Obj *obj)
{
    if (!list) return obj;

    list = get_last_child(list);
    obj->next = NULL;
    obj->prev = list;
    list->next = obj;
    return obj;
}

static hpdf_Obj* get_last_child (hpdf_Obj *obj)
{
    hpdf_Obj *start = obj;
    if (obj == NULL) return NULL;

    while (1) {
        if (obj->next == NULL) 
            return obj;
        obj = obj->next;
        if (obj == start) 
            break;
    }
    /* fix looped link */
    obj = start->next;
    start->next = NULL;
    obj->prev   = NULL;
    return start;


    if (obj->next == NULL)
        return obj;
}

static hpdf_Obj* get_first_child (hpdf_Obj *obj)
{
    hpdf_Obj *start = obj;
    if (obj == NULL) return NULL;

    while (1) {
        if (obj->prev == NULL) 
            return obj;
        obj = obj->prev;
        if (obj == start) 
            break;
    }
    /* fix looped link */
    obj = start->prev;
    start->prev = NULL;
    obj->next   = NULL;
    return start;
}
